//////////////////////////////////////////////////////////
//  Metamorph by Daniel Switkin, metamorph@switkin.com
//  Last revised: 3/30/99
//
//  You may do anything you like with this except for
//  commercial applications. If you like it please drop
//  me a line.
//////////////////////////////////////////////////////////

#include "metamorph.h"

inline vector line2vector(line input) {
	vector v;
	v.x = input.x2 - input.x1;
	v.y = input.y2 - input.y1;
	return v;
}

inline vector points2vector(point p1, point p2) {
	vector v;
	v.x = p1.x - p2.x;
	v.y = p1.y - p2.y;
	return v;
}

inline vector perpendicular(line input) {
	vector v;
	v.y = input.x2 - input.x1;
	v.x = (input.y2 - input.y1) * -1;
	return v;
}

inline double dotproduct(vector v, vector w) {
	double r = (v.x * w.x) + (v.y * w.y);
	return r;
}

inline vector dotproduct(vector v, double s) {
	vector r;
	r.x = v.x * s;
	r.y = v.y * s;
	return r;
}

inline double magnitude(vector v) {
	double m = (v.x * v.x) + (v.y * v.y);
	m = sqrt(m);
	return m;
}

bool morph_forward() {
	for (int count = 0; count < framecount; count++) {
		Targa *current = new Targa();
		char temp[100];
		sprintf(temp, "forward%03d.tga", count);
		if (!current->New(temp, start->width, start->height)) {
			delete current;
			return false;
		}
		printf("morph_forward creating file %s\n", temp);
		
		// Calculate the interpolated lines for this frame of the animtion in advance
		line *interplines = new line[linecount];
		double blend = (double)(count + 1.0) / (double)(framecount + 1.0);
		for (int lc = 0; lc < linecount; lc++) {
			interplines[lc].x1 = (int)(blend * endlines[lc].x1 + (1.0 - blend) * startlines[lc].x1);
			interplines[lc].x2 = (int)(blend * endlines[lc].x2 + (1.0 - blend) * startlines[lc].x2);
			interplines[lc].y1 = (int)(blend * endlines[lc].y1 + (1.0 - blend) * startlines[lc].y1);
			interplines[lc].y2 = (int)(blend * endlines[lc].y2 + (1.0 - blend) * startlines[lc].y2);
		}
		
		// For each pixel...
		for (int y = 0; y < start->height; y++) {
			for (int x = 0; x < start->width; x++) {
				point Dsum = {0, 0};
				double weightsum = 0;
				point X = {x, y};
				// For each line...
				for (int lc = 0; lc < linecount; lc++) {
					// Calculate u and v
					point P = {interplines[lc].x1, interplines[lc].y1};
					double top = dotproduct(points2vector(X, P), line2vector(interplines[lc]));
					double bot = pow(magnitude(line2vector(interplines[lc])), 2);
					double u = top / bot;
					
					top = dotproduct(points2vector(X, P), perpendicular(interplines[lc]));
					bot = pow(magnitude(line2vector(interplines[lc])), 2);
					double v = top / bot;
					
					// Calculate Xprime for this u and v
					vector a = dotproduct(line2vector(startlines[lc]), u);
					vector b = dotproduct(perpendicular(startlines[lc]), v);
					double xp = a.x + b.x + startlines[lc].x1;
					double yp = a.y + b.y + startlines[lc].y1;
					
					// Find displacement of Xprime and distance from X to PQ
					point Di = { (xp - X.x), (yp - X.y) };
					double dist;
					if (u > 0 && u < 1) dist = fabs(v) * magnitude(line2vector(interplines[lc]));
					else if (u <= 0) dist = magnitude(points2vector(X, P));
					else {
						point Q = {interplines[lc].x2, interplines[lc].y2};
						dist = magnitude(points2vector(X, Q));
					}
					
					// Calculate weight and update Dsum and weightsum
					double weight = pow(pow(magnitude(points2vector(X, P)), WEIGHT_P) / (WEIGHT_A + dist), WEIGHT_B);
					Dsum.x += Di.x * weight;
					Dsum.y += Di.y * weight;
					weightsum += weight;
				}
				
				// Find final Xprime
				double xp = X.x + (Dsum.x / weightsum);
				double yp = X.y + (Dsum.y / weightsum);
				
				// Bilinear interpolation
				double i;
				double xfrac = modf(xp, &i);
				int xfloor = (int)i;
				int xceil = xfloor + 1;
				double yfrac = modf(yp, &i);
				int yfloor = (int)i;
				int yceil = yfloor + 1;
				if (xfloor < 0 || yfloor < 0 || xceil >= start->width || yceil >= start->height) {
					current->data[y][x * 3] = BACKGROUND;
					current->data[y][x * 3 + 1] = BACKGROUND;
					current->data[y][x * 3 + 2] = BACKGROUND;
					continue;
				}
				
				double top = (1.0 - xfrac) * start->data[yfloor][xfloor * 3] +
					xfrac * start->data[yfloor][xceil * 3];
				double bot = (1.0 - xfrac) * start->data[yceil][xfloor * 3] +
					xfrac * start->data[yceil][xceil * 3];
				current->data[y][x * 3] = (int)((1.0 - yfrac) * top + yfrac * bot);
				
				top = (1.0 - xfrac) * start->data[yfloor][xfloor * 3 + 1] +
					xfrac * start->data[yfloor][xceil * 3 + 1];
				bot = (1.0 - xfrac) * start->data[yceil][xfloor * 3 + 1] +
					xfrac * start->data[yceil][xceil * 3 + 1];
				current->data[y][x * 3 + 1] = (int)((1.0 - yfrac) * top + yfrac * bot);
				
				top = (1.0 - xfrac) * start->data[yfloor][xfloor * 3 + 2] +
					xfrac * start->data[yfloor][xceil * 3 + 2];
				bot = (1.0 - xfrac) * start->data[yceil][xfloor * 3 + 2] +
					xfrac * start->data[yceil][xceil * 3 + 2];
				current->data[y][x * 3 + 2] = (int)((1.0 - yfrac) * top + yfrac * bot);
			}
		}
		from[count] = current;
		delete interplines;
	}
	return true;
}

bool morph_backward() {
	for (int count = 0; count < framecount; count++) {
		Targa *current = new Targa();
		char temp[100];
		sprintf(temp, "backward%03d.tga", count);
		if (!current->New(temp, end->width, end->height)) {
			delete current;
			return false;
		}
		printf("morph_backward creating file %s\n", temp);
		
		// Calculate the interpolated lines for this frame of the animtion in advance
		line *interplines = new line[linecount];
		double blend = (double)(count + 1.0) / (double)(framecount + 1.0);
		for (int lc = 0; lc < linecount; lc++) {
			interplines[lc].x1 = (int)((1.0 - blend) * endlines[lc].x1 + blend * startlines[lc].x1);
			interplines[lc].x2 = (int)((1.0 - blend) * endlines[lc].x2 + blend * startlines[lc].x2);
			interplines[lc].y1 = (int)((1.0 - blend) * endlines[lc].y1 + blend * startlines[lc].y1);
			interplines[lc].y2 = (int)((1.0 - blend) * endlines[lc].y2 + blend * startlines[lc].y2);
		}
		
		// For each pixel...
		for (int y = 0; y < start->height; y++) {
			for (int x = 0; x < start->width; x++) {
				point Dsum = {0, 0};
				double weightsum = 0;
				point X = {x, y};
				// For each line...
				for (int lc = 0; lc < linecount; lc++) {
					// Calculate u and v
					point P = {interplines[lc].x1, interplines[lc].y1};
					double top = dotproduct(points2vector(X, P), line2vector(interplines[lc]));
					double bot = pow(magnitude(line2vector(interplines[lc])), 2);
					double u = top / bot;
					
					top = dotproduct(points2vector(X, P), perpendicular(interplines[lc]));
					bot = pow(magnitude(line2vector(interplines[lc])), 2);
					double v = top / bot;
					
					// Calculate Xprime for this u and v
					vector a = dotproduct(line2vector(endlines[lc]), u);
					vector b = dotproduct(perpendicular(endlines[lc]), v);
					double xp = a.x + b.x + endlines[lc].x1;
					double yp = a.y + b.y + endlines[lc].y1;
					
					// Find displacement of Xprime and distance from X to PQ
					point Di = { (xp - X.x), (yp - X.y) };
					double dist;
					if (u > 0 && u < 1) dist = fabs(v) * magnitude(line2vector(interplines[lc]));
					else if (u <= 0) dist = magnitude(points2vector(X, P));
					else {
						point Q = {interplines[lc].x2, interplines[lc].y2};
						dist = magnitude(points2vector(X, Q));
					}
					
					// Calculate weight and update Dsum and weightsum
					double weight = pow(pow(magnitude(points2vector(X, P)), WEIGHT_P) / (WEIGHT_A + dist), WEIGHT_B);
					Dsum.x += Di.x * weight;
					Dsum.y += Di.y * weight;
					weightsum += weight;
				}
				
				// Find final Xprime
				double xp = X.x + (Dsum.x / weightsum);
				double yp = X.y + (Dsum.y / weightsum);
				
				// Bilinear interpolation
				double i;
				double xfrac = modf(xp, &i);
				int xfloor = (int)i;
				int xceil = xfloor + 1;
				double yfrac = modf(yp, &i);
				int yfloor = (int)i;
				int yceil = yfloor + 1;
				
				if (xfloor < 0 || yfloor < 0 || xceil >= start->width || yceil >= start->height) {
					current->data[y][x * 3] = BACKGROUND;
					current->data[y][x * 3 + 1] = BACKGROUND;
					current->data[y][x * 3 + 2] = BACKGROUND;
					continue;
				}
				
				double top = (1.0 - xfrac) * end->data[yfloor][xfloor * 3] +
					xfrac * end->data[yfloor][xceil * 3];
				double bot = (1.0 - xfrac) * end->data[yceil][xfloor * 3] +
					xfrac * end->data[yceil][xceil * 3];
				current->data[y][x * 3] = (int)((1.0 - yfrac) * top + yfrac * bot);
				
				top = (1.0 - xfrac) * end->data[yfloor][xfloor * 3 + 1] +
					xfrac * end->data[yfloor][xceil * 3 + 1];
				bot = (1.0 - xfrac) * end->data[yceil][xfloor * 3 + 1] +
					xfrac * end->data[yceil][xceil * 3 + 1];
				current->data[y][x * 3 + 1] = (int)((1.0 - yfrac) * top + yfrac * bot);
				
				top = (1.0 - xfrac) * end->data[yfloor][xfloor * 3 + 2] +
					xfrac * end->data[yfloor][xceil * 3 + 2];
				bot = (1.0 - xfrac) * end->data[yceil][xfloor * 3 + 2] +
					xfrac * end->data[yceil][xceil * 3 + 2];
				current->data[y][x * 3 + 2] = (int)((1.0 - yfrac) * top + yfrac * bot);
			}
		}
		to[count] = current;
		delete interplines;
	}
	return true;
}

// Grab lines from disk and place into startlines and endlines arrays
bool load_lines() {
	FILE *fp;
	if ((fp = fopen(linefile, "r")) == NULL) {
		printf("Failed to open %s\n", linefile);
		return false;
	}
	fscanf(fp, "%d\n", &linecount);
	startlines = new line[linecount];
	endlines = new line[linecount];
	printf("Found %d lines in file %s\n", linecount, linefile);
	
	for (int x = 0; x < linecount; x++) {
		fscanf(fp, "%d%d%d%d%d%d%d%d\n", &(startlines[x].x1),
			&(startlines[x].y1), &(startlines[x].x2), &(startlines[x].y2),
			&(endlines[x].x1), &(endlines[x].y1), &(endlines[x].x2),
			&(endlines[x].y2));
		printf("%d,%d %d,%d --> %d,%d %d,%d\n", startlines[x].x1,
			startlines[x].y1, startlines[x].x2, startlines[x].y2,
			endlines[x].x1, endlines[x].y1, endlines[x].x2,
			endlines[x].y2);
	}
	fclose(fp);
	return true;
}

// Do the cross dissolve after all the frames have been generated
bool dissolve() {
	char filename[255];
	int count = 0;
	int width = from[0]->width;
	int height = from[0]->height;
	
	while (count < framecount) {
		sprintf(filename, "%s%03d.tga", finalfile, count);
		Targa *result = new Targa();
		if (!result->New(filename, width, height)) {
			delete result;
			return false;
		}
		
		double blend = (double)(count + 1.0) / (double)(framecount + 1.0);
		printf("Writing %s with blend %03f\n", filename, blend);

		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
			    result->data[y][x * 3] = (from[count]->data[y][x * 3] * (1.0 - blend)) +
			    	(to[framecount - count - 1]->data[y][x * 3] * blend);
			    result->data[y][x * 3 + 1] = (from[count]->data[y][x * 3 + 1] * (1.0 - blend)) +
			    	(to[framecount - count - 1]->data[y][x * 3 + 1] * blend);
				result->data[y][x * 3 + 2] = (from[count]->data[y][x * 3 + 2] * (1.0 - blend)) +
			    	(to[framecount - count - 1]->data[y][x * 3 + 2] * blend);
			}
		}
		
		if (!result->Save()) {
			delete result;
			return false;
		}
		delete result;
		count++;
	}
	return true;
}

int main(int argc, char **argv) {
	if (argc != 6) {
		printf("Metamorph v1.0 for BeOS by Daniel Switkin, metamorph@switkin.com\n");
		printf("Home page: http://www.switkin.com/software/metamorph\n\n");
		printf("usage: metamorph startimage.tga endimage.tga lines.txt output framecount\n\n");
		printf("where startimage.tga and endimage.tga are the pictures to be morphed,\n");
		printf("as 24 bit targa files, lines.txt are the guidelines (see sample.txt),\n");
		printf("and output is the name of the sequence of images to be written (no suffix)\n");
		return -1;
	}
	
	strcpy(startfile, argv[1]);
	strcpy(endfile, argv[2]);
	strcpy(linefile, argv[3]);
	strcpy(finalfile, argv[4]);
	framecount = atoi(argv[5]);
	
	start = new Targa();
	if (!start->Load(startfile, false)) {
		delete start;
		return -1;
	}
	
	end = new Targa();
	if (!end->Load(endfile, false)) {
		delete start;
		delete end;
		return -1;
	}
	
	if (start->height != end->height || start->width != end->width) {
		printf("Input images are not the same size, so exiting\n");
		delete start;
		delete end;
		return -1;
	}
	
	from = to = NULL;
	from = (Targa **)malloc(framecount * 4);
	to = (Targa **)malloc(framecount * 4);
	if (from == NULL || to == NULL) {
		if (from != NULL) free(from);
		if (to != NULL) free(to);
		delete start;
		delete end;
		return -1;
	}
	
	// Logical OR will stop as soon as any true condition is found - this is just
	// shorthand to make sure memory cleanup happens no matter what
	if (!load_lines() || !morph_forward() || !morph_backward() || !dissolve() || true) {
		for (int x = 0; x < framecount; x++) {
			if (from[x] != NULL) delete from[x];
			if (to[x] != NULL) delete to[x];
		}
		free(from);
		free(to);
		delete start;
		delete end;
		delete startlines;
		delete endlines;
	}
	return 0;
}