//////////////////////////////////////////////////////////
//  Targa library by Daniel Switkin, utilities@switkin.com
//  Last revised: 3/30/99
//
//  You may do anything you like with this but you must
//  leave this comment in the source, and preferably
//  drop me an email if you use it.
//////////////////////////////////////////////////////////

#ifndef TARGA_H
#include "targa.h"
#endif

/* This is called whether you are reading or writing a targa */
Targa::Targa() {
	width = height = 0;
	properties = STANDARD;
	for (int x = 0; x < 18; x++) header[x] = 0;
	data = NULL;
	scanline = NULL;
	fp = NULL;
}

/* This is how to create a new targa - you must call Save() or else
use will lose whatever you put in here */
bool Targa::New(char *name, int width, int height) {
	this->width = width;
	this->height = height;
	strcpy(this->name, name);
	
	/* Create a two dimensional array */
    if ((data = (unsigned char **)malloc(height * 4)) == NULL) {
		fprintf(stdout, "Out of memory\n");
        return false;
    }
	for (int y = 0; y < height; y++) {
		data[y] = new unsigned char[width * 3];
		if (data[y] == NULL) {
			fprintf(stdout, "Out of memory\n");
    	    for (int x = 0; x < y; x++) {
    	    	delete data[x];
    	    }
    	    free(data);
        	return false;
    	}
	}
	
	/* Set all pixels to black */
	for (int y = 0; y < height; y++) {
    	for (int x = 0; x < width * 3; x++) {
    		data[y][x] = 0;
    	}
    }
		
    properties = TOP_DOWN;
    return true;
}

/* Write targa to disk - does not destroy the object */
bool Targa::Save() {
	/* Fill header with dimensions and type */
	header[2] = 2;
	header[12] = width % 256;
	header[13] = width / 256;
	header[14] = height % 256;
	header[15] = height / 256;
	header[16] = 24;
	header[17] = 0x20;
	
	/* Open the file */
	FILE *savefp;
	if ((savefp = fopen(name, "wb")) == NULL) {
		fprintf(stdout, "Failed to open %s\n", name);
		return false;
	}
	
	/* Write header and data */
	fwrite(header, sizeof(header), 1, savefp);
	for (int y = 0; y < height; y++) {
		fwrite(data[y], width * 3, 1, savefp);
	}
	
	fclose(savefp);
	return true;
}

/* Read a targa from disk into either a two dimensional array
or line by line. Elements are blue, green, red regardless of
which mode you use. These are meant to be read-only images. */
bool Targa::Load(char *name, bool linebyline) {
    strcpy(this->name, name);
    if (strstr(name, ".tga") == NULL) strcat(name, ".tga");
    
    /* Open the file */
    if ((fp = fopen(name, "rb")) == NULL) {
        fprintf(stdout, "Failed to open %s\n", name);
        return false;
    }
    /* Read the 18 byte header */
    if (fread(header, 18, 1, fp) != 1) {
        fprintf(stdout, "Cannot read header of %s\n", name);
        fclose(fp);
        return false;
    }
    
    /* Set status info */
    switch (header[2]) {
        case 1:
            /* Uncompressed, color-mapped images */
            properties = properties | PALETTE;
            break;
        case 2:
            /* Uncompressed, RGB images */
            properties = properties | STANDARD;
            break;
        case 9:
            /* Runlength encoded color-mapped images */
            properties = properties | COMPRESSED;
            properties = properties | PALETTE;
            break;
        case 10:
            /* Runlength encoded RGB images */
            properties = properties | COMPRESSED;
            break;
        default:
            /* Other unknown format */
            properties = properties | OTHER;
            break;
    }
    
    switch (header[16]) {
        case 24:
            /* 24 bit image */
            properties = properties | STANDARD;
            break;
        case 32:
            /* 32 bit image */
            properties = properties | ALPHA;
            break;
        default:
            properties = properties | OTHER;
            break;
    }
    
    /* Is the image top down (bit 5 on) */
    if (header[17] & 0x20) properties = properties | TOP_DOWN;
    /* Exit if unusable type */
    if (properties > (TOP_DOWN | COMPRESSED)) {
        fprintf(stdout, "Unrecognized targa format in %s\n", name);
        fclose(fp);
        return false;
    }
    
    /* Get width and height */
    width = header[12] + (header[13] << 8);
    height = header[14] + (header[15] << 8);
    /* Init some stuff */
    compressed = normal = 0;
    r = g = b = 0;
    
    if (linebyline) {
	    /* Malloc the scanline for this image */
	    if ((scanline = (unsigned char *)malloc(width * 3)) == NULL) {
	        fprintf(stdout, "Out of memory\n");
	        fclose(fp);
	        return false;
	    }
	} else {
		/* Malloc the scanline for this image */
		if ((scanline = (unsigned char *)malloc(width * 3)) == NULL) {
	        fprintf(stdout, "Out of memory\n");
	        fclose(fp);
	        return false;
	    }
	    
		/* Allocate a full 2D array for the image */
		if ((data = (unsigned char **)malloc(height * 4)) == NULL) {
			fprintf(stdout, "Out of memory\n");
	        fclose(fp);
	        return false;
	    }
		for (int y = 0; y < height; y++) {
			data[y] = new unsigned char[width * 3];
			if (data[y] == NULL) {
				fprintf(stdout, "Out of memory\n");
	    	    fclose(fp);
	    	    for (int x = 0; x < y; x++) {
	    	    	delete data[x];
	    	    }
	    	    free(data);
	        	return false;
	    	}
		}
	    
	    /* Fill the data array by calling NextScanLine() */
	    if (properties & TOP_DOWN) {
	    	for (int y = 0; y < height; y++) {
	    		NextScanLine();
	    		memcpy(data[y], scanline, width * 3);
	    	}
	    } else {
	    	for (int y = height - 1; y >= 0; y--) {
	    		NextScanLine();
	    		memcpy(data[y], scanline, width * 3);
	    	}
	    }
	}
	return true;
}

/* Can only use this if you Load()ed the image linebyline */
void Targa::NextScanLine() {
    int w, x;
    unsigned char header;

    if (properties & COMPRESSED) {
        w = width;
        while (w>0) {
            /* Add leftover compressed pixels if present */
            for (x=0;(compressed > 0) && (w > 0);x++) {
                scanline[(width - w) * 3] = b;
                scanline[(width - w) * 3 + 1] = g;
                scanline[(width - w) * 3 + 2] = r;
                compressed--;
                w--;
            }
            /* Add leftover normal pixels if present */
            if (normal > 0 && w > 0) {
                x = (normal > w) ? w : normal;
                if(fread((unsigned char *)scanline + ((width - w) * 3), x*3, 1, fp) != 1) {
                    fprintf(stdout, "Premature EOF in %s\n", name);
                    fclose(fp);
                    exit(-1);
                }
                normal -= x;
                w -= x;
            }
            /* Get new packet if necessary */
            if (compressed == 0 && normal == 0 && w > 0) {
                if (fread(&header, 1, 1, fp) != 1) {
                    fprintf(stdout, "Premature EOF in %s at new packet\n", name);
                    fclose(fp);
                    exit(-1);
                }
                if (header & 0x80) {
                    compressed = (header & 0x7f) + 1;
                    b = fgetc(fp);
                    g = fgetc(fp);
                    r = fgetc(fp);
                }
                else normal = (header & 0x7f) + 1;
            }
        }
    }
    /* Uncompressed image, just grab the next scanline */
    else {
        if (fread(scanline, width * 3, 1, fp) != 1) {
            fprintf(stdout, "Premature EOF in %s\n", name);
            fclose(fp);
            exit(-1);
        }
    }
}

/* Destroy targa object - works for both loading modes and new images */
Targa::~Targa() {
	if (fp != NULL) fclose(fp);
    if (scanline != NULL) free(scanline);
    if (data != NULL) {
    	for (int y = 0; y < height; y++) {
    		delete data[y];
    	}
    	free(data);
    }
}